/* ****************************************************** */
/*    Wilson fermion solver                               */
/*                                                        */
/*  C/C++/CUDA (single precision) version                 */
/*                             30 Apr 2009  H.Matsufuru   */
/*                                                        */
/*                     Copyright(c) Hideo Matsufuru 2009  */
/* ****************************************************** */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>

#include "lattsize.h"

#include "cg_wilson_single.h"

#include "cg_wilson.h"

/*
inline int idx(int ivc, int id,
                      int ix, int iy, int iz, int it){
  return ivc + id*Nvc
         + Nvc*ND*(ix + iy*Nx + iz*Nx*Ny + it*Nx*Ny*Nz);
}
*/

__device__
inline int idx(int ivc, int id, int iv){
                         return ivc + Nvc*(id + iv*ND); }

__device__
inline int idg(int ivc, int ic, int iv){
                       return ivc + Nvc*(ic + iv*Ncol); }

__device__
inline int idxv(int ix, int iy, int iz, int it){
         return ix + Nx*(iy + Ny*(iz + it*Nz));  }


/* ****************************************************** */
void test_WilsonD_gpu(double *ud){

  static float u[Ndf*Nst*4];
  static float v1[Nvst];

  for(int i=0; i < Ndf*Nst*4; i++){
      u[i] = (float) ud[i];  }

  for(int i=0; i < Nvst; i++){
      v1[i] = 0.0f; }
  v1[0] = 1.0f;

  int nrepeat = 1000;
  int fop_h = 1368;
      //  (1368 is FPops of H_W per site)
  double fop = ((double) fop_h*Nst) * ((double) nrepeat*2);

  printf("\nPerformance of Wilson D-mult:\n");
  printf(" Number of H_W mult: %8d\n", nrepeat*2);

  // device variables
  float *u_dev;
  float *v1_dev, *v2_dev, *v3_dev;
  cudaMalloc( (void**)&u_dev,Ndf*Nst*4*sizeof(float) );
  cudaMalloc( (void**)&v1_dev,Nvst*sizeof(float) );
  cudaMalloc( (void**)&v2_dev,Nvst*sizeof(float) );
  cudaMalloc( (void**)&v3_dev,Nvst*sizeof(float) );

  cudaMemcpy(u_dev,u,Ndf*Nst*4*sizeof(float),cudaMemcpyHostToDevice);

  int nblock_x = Nz;
  int nblock_y = Nt;
  int nthread_x = Nx;
  int nthread_y = Ny;
  dim3 dimGrid(nblock_x,nblock_y);
  dim3 dimBlock(nthread_x,nthread_y);

  cudaMemcpy(v1_dev,v1,Nvst*sizeof(float),cudaMemcpyHostToDevice);

  clock_t t_init = clock();

  for(int i=0; i<nrepeat; i++){

    opr_H_dev<<< dimGrid, dimBlock >>>(v2_dev,u_dev,v1_dev);
    opr_H_dev<<< dimGrid, dimBlock >>>(v1_dev,u_dev,v2_dev);

   // printf("  i = %6d\n", i);
  }

  cudaThreadSynchronize();

  clock_t t_fin = clock();
  double etime = (double)(t_fin-t_init)/CLOCKS_PER_SEC;

  printf(" time  = %14.6f [sec]\n", etime);
  printf(" FPops = %19.6e\n", fop);
  printf(" GFlops = %19.6f\n", fop*1.e-9/etime);

  cudaMemcpy(v1,v1_dev,Nvst*sizeof(float),cudaMemcpyDeviceToHost);

  cudaFree(u_dev);
  cudaFree(v1_dev);
  cudaFree(v2_dev);
  cudaFree(v3_dev);

}
/* ****************************************************** */
void solve_CG_gpu(double enorm, int &nconv, double &diff,
                       double *xq, double *ud, double *b){

  int niter = 1000;

  static float u[Ndf*Nst*4];
  static float x[Nvst], s[Nvst], r[Nvst], p[Nvst];
  static float v[Nvst];

  for(int i=0; i < Nvst; i++){
      s[i] = (float) b[i];  }

  for(int i=0; i < Ndf*Nst*4; i++){
      u[i] = (float) ud[i];  }

  float snorm, sr;
  sr = 0.0;
  for(int i=0; i < Nvst; i++){
      sr += s[i]*s[i];  }
  snorm = 1.0/sr;
  printf("snorm = %f\n", snorm);


  // device variables
  float *u_dev;
  float *v1_dev, *v2_dev, *v3_dev;
  cudaMalloc( (void**)&u_dev,Ndf*Nst*4*sizeof(float) );
  cudaMalloc( (void**)&v1_dev,Nvst*sizeof(float) );
  cudaMalloc( (void**)&v2_dev,Nvst*sizeof(float) );
  cudaMalloc( (void**)&v3_dev,Nvst*sizeof(float) );

  cudaMemcpy(u_dev,u,Ndf*Nst*4*sizeof(float),cudaMemcpyHostToDevice);

  int nblock_x = Nz;
  int nblock_y = Nt;
  int nthread_x = Nx;
  int nthread_y = Ny;
  dim3 dimGrid(nblock_x,nblock_y);
  dim3 dimBlock(nthread_x,nthread_y);


// =====  solver main part  =====

  float rr, rrp;
  nconv = -1;

// initial step
//  solve_CG_init(rrp, rr, u, x, r, s, p, s);

  //equate(Nvst,r,s);
  //equate(Nvst,x,s);
  for(int i=0; i < Nvst; i++){
      r[i] = s[i];
      x[i] = s[i];
  }

  //opr_DdagD_dev(s,u,x);
  cudaMemcpy(v1_dev,x,Nvst*sizeof(float),cudaMemcpyHostToDevice);
  opr_H_dev<<< dimGrid, dimBlock >>>(v2_dev,u_dev,v1_dev);
  opr_H_dev<<< dimGrid, dimBlock >>>(v3_dev,u_dev,v2_dev);
  cudaMemcpy(s,v3_dev,Nvst*sizeof(float),cudaMemcpyDeviceToHost);

  //selfadd(Nvst,r,s,-1.0);
  //equate(Nvst,p,r);
  rr = 0.0;
  for(int i=0; i < Nvst; i++){
      r[i] += - s[i];
      p[i] = r[i];
      rr += r[i]*r[i];
  }

  //normv(Nvst,rr,r);
  rrp = rr;

  printf("  init: %22.15e\n", rr*snorm);


  for(int iter=0; iter < niter; iter++){

   // ----- CG steps -----

    float cr, bk, pap;

    // opr_DdagD_dev(s,u,p);
    cudaMemcpy(v1_dev,p,Nvst*sizeof(float),cudaMemcpyHostToDevice);
    opr_H_dev<<< dimGrid, dimBlock >>>(v2_dev,u_dev,v1_dev);
    opr_H_dev<<< dimGrid, dimBlock >>>(v3_dev,u_dev,v2_dev);
    cudaMemcpy(s,v3_dev,Nvst*sizeof(float),cudaMemcpyDeviceToHost);

    //vecprd(Nvst,pap,s,p);
    pap = 0.0;
    for(int i=0; i < Nvst; i++){
      pap += s[i]*p[i];  }
    cr = rrp/pap;

    // selfadd(Nvst,x,p,cr);
    // selfadd(Nvst,r,s,-cr);
    // normv(Nvst,rr,r);
    rr = 0.0;
    for(int i=0; i < Nvst; i++){
      x[i] +=   cr * p[i];
      r[i] += - cr * s[i];
      rr += r[i]*r[i];
    }
    bk = rr/rrp;

    // selfsprd(Nvst,p,bk);
    // selfadd(Nvst,p,r,1.0);
    for(int i=0; i < Nvst; i++){
      p[i] =  bk * p[i];
      p[i] += r[i];
    }

    rrp = rr;

    printf("%6d  %22.15e\n", iter, rr*snorm);

    if(rr*snorm < enorm){
      nconv = iter;
      break;
    }

  }

  if(nconv == -1){
    printf(" not converged\n");
  }

  //  printf("converged:\n");
  //  printf("  nconv = %d\n", nconv);


  cudaFree(u_dev);
  cudaFree(v1_dev);
  cudaFree(v2_dev);
  cudaFree(v3_dev);


  for(int i=0; i < Nvst; i++){
      xq[i] = (double) x[i];  }


  static double vt[Nvst];
  opr_DdagD(vt,ud,xq);
  selfadd(Nvst,vt,b,-1.0);
  normv(Nvst,diff,vt);
  
}
/* ****************************************************** */
__global__ void opr_H_dev(float *v, float *u, float *w){

  mult_xp_dev(v,u,w);
  mult_xm_dev(v,u,w);
  mult_yp_dev(v,u,w);
  mult_ym_dev(v,u,w);
  mult_zp_dev(v,u,w);
  mult_zm_dev(v,u,w);
  mult_tp_dev(v,u,w);
  mult_tm_dev(v,u,w);

  int iv = idxv(threadIdx.x,threadIdx.y,blockIdx.x,blockIdx.y);

  for(int ivc=0; ivc < Nvc; ivc++){
     float v3 = v[idx(ivc,2,iv)];
     float v4 = v[idx(ivc,3,iv)];
     v[idx(ivc,2,iv)] = w[idx(ivc,0,iv)] - CKs * v[idx(ivc,0,iv)];
     v[idx(ivc,3,iv)] = w[idx(ivc,1,iv)] - CKs * v[idx(ivc,1,iv)];
     v[idx(ivc,0,iv)] = w[idx(ivc,2,iv)] - CKs * v3;
     v[idx(ivc,1,iv)] = w[idx(ivc,3,iv)] - CKs * v4;
  }


}
/* ****************************************************** */
__device__ void mult_xp_dev(float *v2, float *u, float *v1){

  int idir = 0;

  float vt1[Nvc], vt2[Nvc];
  float wt1r, wt1i, wt2r, wt2i;

  int ivx = idxv(threadIdx.x,threadIdx.y,blockIdx.x,blockIdx.y);
  int nn = (threadIdx.x + 1) % Nx;
  int ivn = idxv(nn,threadIdx.y,blockIdx.x,blockIdx.y);
  int ivg = ivx + Nst*idir;

  for(int ic=0; ic < Ncol; ic++){
    vt1[2*ic  ] = v1[idx(2*ic  ,0,ivn)] - v1[idx(2*ic+1,3,ivn)];
    vt1[2*ic+1] = v1[idx(2*ic+1,0,ivn)] + v1[idx(2*ic  ,3,ivn)];
    vt2[2*ic  ] = v1[idx(2*ic  ,1,ivn)] - v1[idx(2*ic+1,2,ivn)];
    vt2[2*ic+1] = v1[idx(2*ic+1,1,ivn)] + v1[idx(2*ic  ,2,ivn)];
  }

  for(int ic=0; ic < Ncol; ic++){

    wt1r = u[idg(0,ic,ivg)]*vt1[0] - u[idg(1,ic,ivg)]*vt1[1]
         + u[idg(2,ic,ivg)]*vt1[2] - u[idg(3,ic,ivg)]*vt1[3]
         + u[idg(4,ic,ivg)]*vt1[4] - u[idg(5,ic,ivg)]*vt1[5];
    wt1i = u[idg(0,ic,ivg)]*vt1[1] + u[idg(1,ic,ivg)]*vt1[0]
         + u[idg(2,ic,ivg)]*vt1[3] + u[idg(3,ic,ivg)]*vt1[2]
         + u[idg(4,ic,ivg)]*vt1[5] + u[idg(5,ic,ivg)]*vt1[4];

    wt2r = u[idg(0,ic,ivg)]*vt2[0] - u[idg(1,ic,ivg)]*vt2[1]
         + u[idg(2,ic,ivg)]*vt2[2] - u[idg(3,ic,ivg)]*vt2[3]
         + u[idg(4,ic,ivg)]*vt2[4] - u[idg(5,ic,ivg)]*vt2[5];
    wt2i = u[idg(0,ic,ivg)]*vt2[1] + u[idg(1,ic,ivg)]*vt2[0]
         + u[idg(2,ic,ivg)]*vt2[3] + u[idg(3,ic,ivg)]*vt2[2]
         + u[idg(4,ic,ivg)]*vt2[5] + u[idg(5,ic,ivg)]*vt2[4];

    v2[idx(2*ic  ,0,ivx)] =  wt1r;
    v2[idx(2*ic+1,0,ivx)] =  wt1i;
    v2[idx(2*ic  ,1,ivx)] =  wt2r;
    v2[idx(2*ic+1,1,ivx)] =  wt2i;
    v2[idx(2*ic  ,2,ivx)] =  wt2i;
    v2[idx(2*ic+1,2,ivx)] = -wt2r;
    v2[idx(2*ic  ,3,ivx)] =  wt1i;
    v2[idx(2*ic+1,3,ivx)] = -wt1r;
  }

}
/* ****************************************************** */
__device__ void mult_xm_dev(float *v2, float *u, float *v1){

  int idir = 0;

  int ivx = idxv(threadIdx.x,threadIdx.y,blockIdx.x,blockIdx.y);
  int nn = (threadIdx.x - 1 + blockDim.x) % Nx;
  int ivn = idxv(nn,threadIdx.y,blockIdx.x,blockIdx.y);
  int ivg = ivn + Nst*idir;

  float vt1[Nvc], vt2[Nvc];
  float wt1r, wt1i, wt2r, wt2i;

  for(int ic=0; ic < Ncol; ic++){
    vt1[2*ic  ] = v1[idx(2*ic  ,0,ivn)] + v1[idx(2*ic+1,3,ivn)];
    vt1[2*ic+1] = v1[idx(2*ic+1,0,ivn)] - v1[idx(2*ic  ,3,ivn)];
    vt2[2*ic  ] = v1[idx(2*ic  ,1,ivn)] + v1[idx(2*ic+1,2,ivn)];
    vt2[2*ic+1] = v1[idx(2*ic+1,1,ivn)] - v1[idx(2*ic  ,2,ivn)];
  }

  for(int ic=0; ic < Ncol; ic++){
    int icr = 2*ic;
    int ici = 2*ic+1;

    wt1r = u[idg(icr,0,ivg)]*vt1[0] + u[idg(ici,0,ivg)]*vt1[1]
         + u[idg(icr,1,ivg)]*vt1[2] + u[idg(ici,1,ivg)]*vt1[3]
         + u[idg(icr,2,ivg)]*vt1[4] + u[idg(ici,2,ivg)]*vt1[5];
    wt1i = u[idg(icr,0,ivg)]*vt1[1] - u[idg(ici,0,ivg)]*vt1[0]
         + u[idg(icr,1,ivg)]*vt1[3] - u[idg(ici,1,ivg)]*vt1[2]
         + u[idg(icr,2,ivg)]*vt1[5] - u[idg(ici,2,ivg)]*vt1[4];

    wt2r = u[idg(icr,0,ivg)]*vt2[0] + u[idg(ici,0,ivg)]*vt2[1]
         + u[idg(icr,1,ivg)]*vt2[2] + u[idg(ici,1,ivg)]*vt2[3]
         + u[idg(icr,2,ivg)]*vt2[4] + u[idg(ici,2,ivg)]*vt2[5];
    wt2i = u[idg(icr,0,ivg)]*vt2[1] - u[idg(ici,0,ivg)]*vt2[0]
         + u[idg(icr,1,ivg)]*vt2[3] - u[idg(ici,1,ivg)]*vt2[2]
         + u[idg(icr,2,ivg)]*vt2[5] - u[idg(ici,2,ivg)]*vt2[4];

    v2[idx(2*ic  ,0,ivx)] +=  wt1r;
    v2[idx(2*ic+1,0,ivx)] +=  wt1i;
    v2[idx(2*ic  ,1,ivx)] +=  wt2r;
    v2[idx(2*ic+1,1,ivx)] +=  wt2i;
    v2[idx(2*ic  ,2,ivx)] += -wt2i;
    v2[idx(2*ic+1,2,ivx)] +=  wt2r;
    v2[idx(2*ic  ,3,ivx)] += -wt1i;
    v2[idx(2*ic+1,3,ivx)] +=  wt1r;
  }

}
/* ****************************************************** */
__device__ void mult_yp_dev(float *v2, float *u, float *v1){

  int idir = 1;

  int ivx = idxv(threadIdx.x,threadIdx.y,blockIdx.x,blockIdx.y);
  int nn = (threadIdx.y + 1) % Ny;
  int ivn = idxv(threadIdx.x,nn,blockIdx.x,blockIdx.y);
  int ivg = ivx + Nst*idir;

  float vt1[Nvc], vt2[Nvc];
  float wt1r, wt1i, wt2r, wt2i;

  for(int ic=0; ic < Ncol; ic++){
    vt1[2*ic  ] = v1[idx(2*ic  ,0,ivn)] + v1[idx(2*ic  ,3,ivn)];
    vt1[2*ic+1] = v1[idx(2*ic+1,0,ivn)] + v1[idx(2*ic+1,3,ivn)];
    vt2[2*ic  ] = v1[idx(2*ic  ,1,ivn)] - v1[idx(2*ic  ,2,ivn)];
    vt2[2*ic+1] = v1[idx(2*ic+1,1,ivn)] - v1[idx(2*ic+1,2,ivn)];
  }

  for(int ic=0; ic < Ncol; ic++){

    wt1r = u[idg(0,ic,ivg)]*vt1[0] - u[idg(1,ic,ivg)]*vt1[1]
         + u[idg(2,ic,ivg)]*vt1[2] - u[idg(3,ic,ivg)]*vt1[3]
         + u[idg(4,ic,ivg)]*vt1[4] - u[idg(5,ic,ivg)]*vt1[5];
    wt1i = u[idg(0,ic,ivg)]*vt1[1] + u[idg(1,ic,ivg)]*vt1[0]
         + u[idg(2,ic,ivg)]*vt1[3] + u[idg(3,ic,ivg)]*vt1[2]
         + u[idg(4,ic,ivg)]*vt1[5] + u[idg(5,ic,ivg)]*vt1[4];

    wt2r = u[idg(0,ic,ivg)]*vt2[0] - u[idg(1,ic,ivg)]*vt2[1]
         + u[idg(2,ic,ivg)]*vt2[2] - u[idg(3,ic,ivg)]*vt2[3]
         + u[idg(4,ic,ivg)]*vt2[4] - u[idg(5,ic,ivg)]*vt2[5];
    wt2i = u[idg(0,ic,ivg)]*vt2[1] + u[idg(1,ic,ivg)]*vt2[0]
         + u[idg(2,ic,ivg)]*vt2[3] + u[idg(3,ic,ivg)]*vt2[2]
         + u[idg(4,ic,ivg)]*vt2[5] + u[idg(5,ic,ivg)]*vt2[4];

    v2[idx(2*ic  ,0,ivx)] +=  wt1r;
    v2[idx(2*ic+1,0,ivx)] +=  wt1i;
    v2[idx(2*ic  ,1,ivx)] +=  wt2r;
    v2[idx(2*ic+1,1,ivx)] +=  wt2i;
    v2[idx(2*ic  ,2,ivx)] += -wt2r;
    v2[idx(2*ic+1,2,ivx)] += -wt2i;
    v2[idx(2*ic  ,3,ivx)] +=  wt1r;
    v2[idx(2*ic+1,3,ivx)] +=  wt1i;
  }

}
/* ****************************************************** */
__device__ void mult_ym_dev(float *v2, float *u, float *v1){

  int idir = 1;

  int ivx = idxv(threadIdx.x,threadIdx.y,blockIdx.x,blockIdx.y);
  int nn = (threadIdx.y - 1 + blockDim.y) % Nx;
  int ivn = idxv(threadIdx.x,nn,blockIdx.x,blockIdx.y);
  int ivg = ivn + Nst*idir;

  float vt1[Nvc], vt2[Nvc];
  float wt1r, wt1i, wt2r, wt2i;

  for(int ic=0; ic < Ncol; ic++){
    vt1[2*ic  ] = v1[idx(2*ic  ,0,ivn)] - v1[idx(2*ic  ,3,ivn)];
    vt1[2*ic+1] = v1[idx(2*ic+1,0,ivn)] - v1[idx(2*ic+1,3,ivn)];
    vt2[2*ic  ] = v1[idx(2*ic  ,1,ivn)] + v1[idx(2*ic  ,2,ivn)];
    vt2[2*ic+1] = v1[idx(2*ic+1,1,ivn)] + v1[idx(2*ic+1,2,ivn)];
  }

  for(int ic=0; ic < Ncol; ic++){
    int icr = 2*ic;
    int ici = 2*ic+1;

    wt1r = u[idg(icr,0,ivg)]*vt1[0] + u[idg(ici,0,ivg)]*vt1[1]
         + u[idg(icr,1,ivg)]*vt1[2] + u[idg(ici,1,ivg)]*vt1[3]
         + u[idg(icr,2,ivg)]*vt1[4] + u[idg(ici,2,ivg)]*vt1[5];
    wt1i = u[idg(icr,0,ivg)]*vt1[1] - u[idg(ici,0,ivg)]*vt1[0]
         + u[idg(icr,1,ivg)]*vt1[3] - u[idg(ici,1,ivg)]*vt1[2]
         + u[idg(icr,2,ivg)]*vt1[5] - u[idg(ici,2,ivg)]*vt1[4];

    wt2r = u[idg(icr,0,ivg)]*vt2[0] + u[idg(ici,0,ivg)]*vt2[1]
         + u[idg(icr,1,ivg)]*vt2[2] + u[idg(ici,1,ivg)]*vt2[3]
         + u[idg(icr,2,ivg)]*vt2[4] + u[idg(ici,2,ivg)]*vt2[5];
    wt2i = u[idg(icr,0,ivg)]*vt2[1] - u[idg(ici,0,ivg)]*vt2[0]
         + u[idg(icr,1,ivg)]*vt2[3] - u[idg(ici,1,ivg)]*vt2[2]
         + u[idg(icr,2,ivg)]*vt2[5] - u[idg(ici,2,ivg)]*vt2[4];

    v2[idx(2*ic  ,0,ivx)] +=  wt1r;
    v2[idx(2*ic+1,0,ivx)] +=  wt1i;
    v2[idx(2*ic  ,1,ivx)] +=  wt2r;
    v2[idx(2*ic+1,1,ivx)] +=  wt2i;
    v2[idx(2*ic  ,2,ivx)] +=  wt2r;
    v2[idx(2*ic+1,2,ivx)] +=  wt2i;
    v2[idx(2*ic  ,3,ivx)] += -wt1r;
    v2[idx(2*ic+1,3,ivx)] += -wt1i;
  }

}
/* ****************************************************** */
__device__ void mult_zp_dev(float *v2, float *u, float *v1){

  int idir = 2;

  int ivx = idxv(threadIdx.x,threadIdx.y,blockIdx.x,blockIdx.y);
  int nn = (blockIdx.x + 1) % Nz;
  int ivn = idxv(threadIdx.x,threadIdx.y,nn,blockIdx.y);
  int ivg = ivx + Nst*idir;

  float vt1[Nvc], vt2[Nvc];
  float wt1r, wt1i, wt2r, wt2i;

  for(int ic=0; ic < Ncol; ic++){
    vt1[2*ic  ] = v1[idx(2*ic  ,0,ivn)] - v1[idx(2*ic+1,2,ivn)];
    vt1[2*ic+1] = v1[idx(2*ic+1,0,ivn)] + v1[idx(2*ic  ,2,ivn)];
    vt2[2*ic  ] = v1[idx(2*ic  ,1,ivn)] + v1[idx(2*ic+1,3,ivn)];
    vt2[2*ic+1] = v1[idx(2*ic+1,1,ivn)] - v1[idx(2*ic  ,3,ivn)];
  }

  for(int ic=0; ic < Ncol; ic++){

    wt1r = u[idg(0,ic,ivg)]*vt1[0] - u[idg(1,ic,ivg)]*vt1[1]
         + u[idg(2,ic,ivg)]*vt1[2] - u[idg(3,ic,ivg)]*vt1[3]
         + u[idg(4,ic,ivg)]*vt1[4] - u[idg(5,ic,ivg)]*vt1[5];
    wt1i = u[idg(0,ic,ivg)]*vt1[1] + u[idg(1,ic,ivg)]*vt1[0]
         + u[idg(2,ic,ivg)]*vt1[3] + u[idg(3,ic,ivg)]*vt1[2]
         + u[idg(4,ic,ivg)]*vt1[5] + u[idg(5,ic,ivg)]*vt1[4];

    wt2r = u[idg(0,ic,ivg)]*vt2[0] - u[idg(1,ic,ivg)]*vt2[1]
         + u[idg(2,ic,ivg)]*vt2[2] - u[idg(3,ic,ivg)]*vt2[3]
         + u[idg(4,ic,ivg)]*vt2[4] - u[idg(5,ic,ivg)]*vt2[5];
    wt2i = u[idg(0,ic,ivg)]*vt2[1] + u[idg(1,ic,ivg)]*vt2[0]
         + u[idg(2,ic,ivg)]*vt2[3] + u[idg(3,ic,ivg)]*vt2[2]
         + u[idg(4,ic,ivg)]*vt2[5] + u[idg(5,ic,ivg)]*vt2[4];

    v2[idx(2*ic  ,0,ivx)] +=  wt1r;
    v2[idx(2*ic+1,0,ivx)] +=  wt1i;
    v2[idx(2*ic  ,1,ivx)] +=  wt2r;
    v2[idx(2*ic+1,1,ivx)] +=  wt2i;
    v2[idx(2*ic  ,2,ivx)] +=  wt1i;
    v2[idx(2*ic+1,2,ivx)] += -wt1r;
    v2[idx(2*ic  ,3,ivx)] += -wt2i;
    v2[idx(2*ic+1,3,ivx)] +=  wt2r;
  }

}
/* ****************************************************** */
__device__ void mult_zm_dev(float *v2, float *u, float *v1){

  int idir = 2;

  int ivx = idxv(threadIdx.x,threadIdx.y,blockIdx.x,blockIdx.y);
  int nn = (blockIdx.x - 1 + gridDim.x) % Nz;
  int ivn = idxv(threadIdx.x,threadIdx.y,nn,blockIdx.y);
  int ivg = ivn + Nst*idir;

  float vt1[Nvc], vt2[Nvc];
  float wt1r, wt1i, wt2r, wt2i;

  for(int ic=0; ic < Ncol; ic++){
    vt1[2*ic  ] = v1[idx(2*ic  ,0,ivn)] + v1[idx(2*ic+1,2,ivn)];
    vt1[2*ic+1] = v1[idx(2*ic+1,0,ivn)] - v1[idx(2*ic  ,2,ivn)];
    vt2[2*ic  ] = v1[idx(2*ic  ,1,ivn)] - v1[idx(2*ic+1,3,ivn)];
    vt2[2*ic+1] = v1[idx(2*ic+1,1,ivn)] + v1[idx(2*ic  ,3,ivn)];
  }

  for(int ic=0; ic < Ncol; ic++){
    int icr = 2*ic;
    int ici = 2*ic+1;

    wt1r = u[idg(icr,0,ivg)]*vt1[0] + u[idg(ici,0,ivg)]*vt1[1]
         + u[idg(icr,1,ivg)]*vt1[2] + u[idg(ici,1,ivg)]*vt1[3]
         + u[idg(icr,2,ivg)]*vt1[4] + u[idg(ici,2,ivg)]*vt1[5];
    wt1i = u[idg(icr,0,ivg)]*vt1[1] - u[idg(ici,0,ivg)]*vt1[0]
         + u[idg(icr,1,ivg)]*vt1[3] - u[idg(ici,1,ivg)]*vt1[2]
         + u[idg(icr,2,ivg)]*vt1[5] - u[idg(ici,2,ivg)]*vt1[4];

    wt2r = u[idg(icr,0,ivg)]*vt2[0] + u[idg(ici,0,ivg)]*vt2[1]
         + u[idg(icr,1,ivg)]*vt2[2] + u[idg(ici,1,ivg)]*vt2[3]
         + u[idg(icr,2,ivg)]*vt2[4] + u[idg(ici,2,ivg)]*vt2[5];
    wt2i = u[idg(icr,0,ivg)]*vt2[1] - u[idg(ici,0,ivg)]*vt2[0]
         + u[idg(icr,1,ivg)]*vt2[3] - u[idg(ici,1,ivg)]*vt2[2]
         + u[idg(icr,2,ivg)]*vt2[5] - u[idg(ici,2,ivg)]*vt2[4];

    v2[idx(2*ic  ,0,ivx)] +=  wt1r;
    v2[idx(2*ic+1,0,ivx)] +=  wt1i;
    v2[idx(2*ic  ,1,ivx)] +=  wt2r;
    v2[idx(2*ic+1,1,ivx)] +=  wt2i;
    v2[idx(2*ic  ,2,ivx)] += -wt1i;
    v2[idx(2*ic+1,2,ivx)] +=  wt1r;
    v2[idx(2*ic  ,3,ivx)] +=  wt2i;
    v2[idx(2*ic+1,3,ivx)] += -wt2r;
  }

}
/* ****************************************************** */
__device__ void mult_tp_dev(float *v2, float *u, float *v1){

  int idir = 3;

  int ivx = idxv(threadIdx.x,threadIdx.y,blockIdx.x,blockIdx.y);
  int nn = (blockIdx.y + 1) % Nt;
  int ivn = idxv(threadIdx.x,threadIdx.y,blockIdx.x,nn);
  int ivg = ivx + Nst*idir;

  float vt1[Nvc], vt2[Nvc];
  float wt1r, wt1i, wt2r, wt2i;


  for(int ic=0; ic < Ncol; ic++){
    vt1[2*ic  ] = 2.0f * v1[idx(2*ic  ,2,ivn)];
    vt1[2*ic+1] = 2.0f * v1[idx(2*ic+1,2,ivn)];
    vt2[2*ic  ] = 2.0f * v1[idx(2*ic  ,3,ivn)];
    vt2[2*ic+1] = 2.0f * v1[idx(2*ic+1,3,ivn)];
  }

  for(int ic=0; ic < Ncol; ic++){

    wt1r = u[idg(0,ic,ivg)]*vt1[0] - u[idg(1,ic,ivg)]*vt1[1]
         + u[idg(2,ic,ivg)]*vt1[2] - u[idg(3,ic,ivg)]*vt1[3]
         + u[idg(4,ic,ivg)]*vt1[4] - u[idg(5,ic,ivg)]*vt1[5];
    wt1i = u[idg(0,ic,ivg)]*vt1[1] + u[idg(1,ic,ivg)]*vt1[0]
         + u[idg(2,ic,ivg)]*vt1[3] + u[idg(3,ic,ivg)]*vt1[2]
         + u[idg(4,ic,ivg)]*vt1[5] + u[idg(5,ic,ivg)]*vt1[4];

    wt2r = u[idg(0,ic,ivg)]*vt2[0] - u[idg(1,ic,ivg)]*vt2[1]
         + u[idg(2,ic,ivg)]*vt2[2] - u[idg(3,ic,ivg)]*vt2[3]
         + u[idg(4,ic,ivg)]*vt2[4] - u[idg(5,ic,ivg)]*vt2[5];
    wt2i = u[idg(0,ic,ivg)]*vt2[1] + u[idg(1,ic,ivg)]*vt2[0]
         + u[idg(2,ic,ivg)]*vt2[3] + u[idg(3,ic,ivg)]*vt2[2]
         + u[idg(4,ic,ivg)]*vt2[5] + u[idg(5,ic,ivg)]*vt2[4];

    v2[idx(2*ic  ,2,ivx)] +=  wt1r;
    v2[idx(2*ic+1,2,ivx)] +=  wt1i;
    v2[idx(2*ic  ,3,ivx)] +=  wt2r;
    v2[idx(2*ic+1,3,ivx)] +=  wt2i;
  }

}
/* ****************************************************** */
__device__ void mult_tm_dev(float *v2, float *u, float *v1){

  int idir = 3;

  int ivx = idxv(threadIdx.x,threadIdx.y,blockIdx.x,blockIdx.y);
  int nn = (blockIdx.y - 1 + gridDim.y) % Nt;
  int ivn = idxv(threadIdx.x,threadIdx.y,blockIdx.x,nn);
  int ivg = ivn + Nst*idir;

  float vt1[Nvc], vt2[Nvc];
  float wt1r, wt1i, wt2r, wt2i;

  for(int ic=0; ic < Ncol; ic++){
    vt1[2*ic  ] = 2.0f * v1[idx(2*ic  ,0,ivn)];
    vt1[2*ic+1] = 2.0f * v1[idx(2*ic+1,0,ivn)];
    vt2[2*ic  ] = 2.0f * v1[idx(2*ic  ,1,ivn)];
    vt2[2*ic+1] = 2.0f * v1[idx(2*ic+1,1,ivn)];
  }

  for(int ic=0; ic < Ncol; ic++){
    int icr = 2*ic;
    int ici = 2*ic+1;

    wt1r = u[idg(icr,0,ivg)]*vt1[0] + u[idg(ici,0,ivg)]*vt1[1]
         + u[idg(icr,1,ivg)]*vt1[2] + u[idg(ici,1,ivg)]*vt1[3]
         + u[idg(icr,2,ivg)]*vt1[4] + u[idg(ici,2,ivg)]*vt1[5];
    wt1i = u[idg(icr,0,ivg)]*vt1[1] - u[idg(ici,0,ivg)]*vt1[0]
         + u[idg(icr,1,ivg)]*vt1[3] - u[idg(ici,1,ivg)]*vt1[2]
         + u[idg(icr,2,ivg)]*vt1[5] - u[idg(ici,2,ivg)]*vt1[4];

    wt2r = u[idg(icr,0,ivg)]*vt2[0] + u[idg(ici,0,ivg)]*vt2[1]
         + u[idg(icr,1,ivg)]*vt2[2] + u[idg(ici,1,ivg)]*vt2[3]
         + u[idg(icr,2,ivg)]*vt2[4] + u[idg(ici,2,ivg)]*vt2[5];
    wt2i = u[idg(icr,0,ivg)]*vt2[1] - u[idg(ici,0,ivg)]*vt2[0]
         + u[idg(icr,1,ivg)]*vt2[3] - u[idg(ici,1,ivg)]*vt2[2]
         + u[idg(icr,2,ivg)]*vt2[5] - u[idg(ici,2,ivg)]*vt2[4];

    v2[idx(2*ic  ,0,ivx)] +=  wt1r;
    v2[idx(2*ic+1,0,ivx)] +=  wt1i;
    v2[idx(2*ic  ,1,ivx)] +=  wt2r;
    v2[idx(2*ic+1,1,ivx)] +=  wt2i;

  }

}
/* ****************************************************** */
