/* ****************************************************** */
/*    Library routines for using CUDA                     */
/*                                                        */
/*                             30 Apr 2009  H.Matsufuru   */
/*                                                        */
/*                     Copyright(c) Hideo Matsufuru 2009  */
/* ****************************************************** */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "lattsize.h"

#include "lib_cuda.h"

/* ****************************************************** */
void cuda_init(){

int ndev;
cudaGetDeviceCount(&ndev);
printf("Number of devices = %d\n", ndev);

int idev = 2;

struct cudaDeviceProp prop;

cudaGetDeviceProperties(&prop,idev);

printf("\nDevice information: idev = %d\n", idev);
printf("Device name:             %s\n", prop.name);
printf("Global memory [B]:       %u\n", prop.totalGlobalMem);
printf("Shared memory/block [B]: %u\n", prop.sharedMemPerBlock);
printf("Register/block[B]:       %d\n", prop.regsPerBlock);
printf("Warp size:               %d\n", prop.warpSize);
printf("Maximum pitch:           %u\n", prop.memPitch);
printf("Maximum threads/block:   %d\n", prop.maxThreadsPerBlock);

printf("Max thread dim:          (%d, %d, %d)\n",
                                        prop.maxThreadsDim[0],
                                        prop.maxThreadsDim[1],
                                        prop.maxThreadsDim[2]);
printf("Max grid size:           (%d, %d, %d)\n",
                                        prop.maxGridSize[0],
                                        prop.maxGridSize[1],
                                        prop.maxGridSize[2]);
printf("Total const memory [B]:  %u\n", prop.totalConstMem);
printf("Compute capability(major):  %d\n", prop.major);
printf("Compute capability(minor):  %d\n", prop.minor);
printf("Clock rate:              %d\n", prop.clockRate);
printf("Texture alignment:       %d\n\n", prop.textureAlignment);

cudaSetDevice(idev);

int idev2;
cudaGetDevice(&idev2);
printf("Used device number = %d\n", idev2);

}
//* ****************************************************** */
